"""Provide the class Message and its subclasses."""


class Message:
    message  = ''
    msg_args = ()

    def __init__(self, filename, loc):
        self.file = filename
        self.line = loc.lineno
        self.col  = loc.col_offset

    def __str__(self):
        return '{}:{}:{}: {}'.format(self.file, self.line,
               self.col+1, self.message % self.msg_args)


class UnusedImport(Message):
    message = '%r imported but unused'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (name,)


class RedefinedWhileUnused(Message):
    message = 'redefinition of unused %r from line %r'

    def __init__(self, filename, loc, name, orig_loc):
        Message.__init__(self, filename, loc)
        self.msg_args = (name, orig_loc.lineno)


class ImportShadowedByLoopVar(Message):
    message = 'import %r from line %r shadowed by loop variable'

    def __init__(self, filename, loc, name, orig_loc):
        Message.__init__(self, filename, loc)
        self.msg_args = (name, orig_loc.lineno)


class ImportStarNotPermitted(Message):
    message = "'from %s import *' only allowed at module level"

    def __init__(self, filename, loc, modname):
        Message.__init__(self, filename, loc)
        self.msg_args = (modname,)


class ImportStarUsed(Message):
    message = "'from %s import *' used; unable to detect undefined names"

    def __init__(self, filename, loc, modname):
        Message.__init__(self, filename, loc)
        self.msg_args = (modname,)


class ImportStarUsage(Message):
    message = "%r may be undefined, or defined from star imports: %s"

    def __init__(self, filename, loc, name, from_list):
        Message.__init__(self, filename, loc)
        self.msg_args = (name, from_list)


class UndefinedName(Message):
    message = 'undefined name %r'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (name,)


class DoctestSyntaxError(Message):
    message = 'syntax error in doctest'

    def __init__(self, filename, loc, position=None):
        Message.__init__(self, filename, loc)
        if position:
            (self.line, self.col) = position
        self.msg_args = ()


class UndefinedExport(Message):
    message = 'undefined name %r in __all__'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (name,)


class UndefinedLocal(Message):
    message = 'local variable %r {0} referenced before assignment'

    default = 'defined in enclosing scope on line %r'
    builtin = 'defined as a builtin'

    def __init__(self, filename, loc, name, orig_loc):
        Message.__init__(self, filename, loc)
        if orig_loc is None:
            self.message = self.message.format(self.builtin)
            self.msg_args = name
        else:
            self.message = self.message.format(self.default)
            self.msg_args = (name, orig_loc.lineno)


class DuplicateArgument(Message):
    message = 'duplicate argument %r in function definition'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (name,)


class MultiValueRepeatedKeyLiteral(Message):
    message = 'dictionary key %r repeated with different values'

    def __init__(self, filename, loc, key):
        Message.__init__(self, filename, loc)
        self.msg_args = (key,)


class MultiValueRepeatedKeyVariable(Message):
    message = 'dictionary key variable %s repeated with different values'

    def __init__(self, filename, loc, key):
        Message.__init__(self, filename, loc)
        self.msg_args = (key,)


class LateFutureImport(Message):
    message = 'from __future__ imports must occur at the beginning of the file'


class FutureFeatureNotDefined(Message):
    """An undefined __future__ feature name was imported."""
    message = 'future feature %s is not defined'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (name,)


class UnusedVariable(Message):
    """
    Indicates that a variable has been explicitly assigned to but not actually
    used.
    """
    message = 'local variable %r is assigned to but never used'

    def __init__(self, filename, loc, names):
        Message.__init__(self, filename, loc)
        self.msg_args = (names,)


class UnusedAnnotation(Message):
    """
    Indicates that a variable has been explicitly annotated to but not actually
    used.
    """
    message = 'local variable %r is annotated but never used'

    def __init__(self, filename, loc, names):
        Message.__init__(self, filename, loc)
        self.msg_args = (names,)


class UnusedIndirectAssignment(Message):
    """A `global` or `nonlocal` statement where the name is never reassigned"""
    message = '`%s %s` is unused: name is never assigned in scope'

    def __init__(self, filename, loc, name):
        Message.__init__(self, filename, loc)
        self.msg_args = (type(loc).__name__.lower(), name)


class ReturnOutsideFunction(Message):
    """
    Indicates a return statement outside of a function/method.
    """
    message = '\'return\' outside function'


class YieldOutsideFunction(Message):
    """
    Indicates a yield or yield from statement outside of a function/method.
    """
    message = '\'yield\' outside function'


# For whatever reason, Python gives different error messages for these two. We
# match the Python error message exactly.
class ContinueOutsideLoop(Message):
    """
    Indicates a continue statement outside of a while or for loop.
    """
    message = '\'continue\' not properly in loop'


class BreakOutsideLoop(Message):
    """
    Indicates a break statement outside of a while or for loop.
    """
    message = '\'break\' outside loop'


class DefaultExceptNotLast(Message):
    """
    Indicates an except: block as not the last exception handler.
    """
    message = 'default \'except:\' must be last'


class TwoStarredExpressions(Message):
    """
    Two or more starred expressions in an assignment (a, *b, *c = d).
    """
    message = 'two starred expressions in assignment'


class TooManyExpressionsInStarredAssignment(Message):
    """
    Too many expressions in an assignment with star-unpacking
    """
    message = 'too many expressions in star-unpacking assignment'


class IfTuple(Message):
    """
    Conditional test is a non-empty tuple literal, which are always True.
    """
    message = '\'if tuple literal\' is always true, perhaps remove accidental comma?'


class AssertTuple(Message):
    """
    Assertion test is a non-empty tuple literal, which are always True.
    """
    message = 'assertion is always true, perhaps remove parentheses?'


class ForwardAnnotationSyntaxError(Message):
    message = 'syntax error in forward annotation %r'

    def __init__(self, filename, loc, annotation):
        Message.__init__(self, filename, loc)
        self.msg_args = (annotation,)


class RaiseNotImplemented(Message):
    message = "'raise NotImplemented' should be 'raise NotImplementedError'"


class InvalidPrintSyntax(Message):
    message = 'use of >> is invalid with print function'


class IsLiteral(Message):
    message = 'use ==/!= to compare constant literals (str, bytes, int, float, tuple)'


class FStringMissingPlaceholders(Message):
    message = 'f-string is missing placeholders'


class TStringMissingPlaceholders(Message):
    message = 't-string is missing placeholders'


class StringDotFormatExtraPositionalArguments(Message):
    message = "'...'.format(...) has unused arguments at position(s): %s"

    def __init__(self, filename, loc, extra_positions):
        Message.__init__(self, filename, loc)
        self.msg_args = (extra_positions,)


class StringDotFormatExtraNamedArguments(Message):
    message = "'...'.format(...) has unused named argument(s): %s"

    def __init__(self, filename, loc, extra_keywords):
        Message.__init__(self, filename, loc)
        self.msg_args = (extra_keywords,)


class StringDotFormatMissingArgument(Message):
    message = "'...'.format(...) is missing argument(s) for placeholder(s): %s"

    def __init__(self, filename, loc, missing_arguments):
        Message.__init__(self, filename, loc)
        self.msg_args = (missing_arguments,)


class StringDotFormatMixingAutomatic(Message):
    message = "'...'.format(...) mixes automatic and manual numbering"


class StringDotFormatInvalidFormat(Message):
    message = "'...'.format(...) has invalid format string: %s"

    def __init__(self, filename, loc, error):
        Message.__init__(self, filename, loc)
        self.msg_args = (error,)


class PercentFormatInvalidFormat(Message):
    message = "'...' %% ... has invalid format string: %s"

    def __init__(self, filename, loc, error):
        Message.__init__(self, filename, loc)
        self.msg_args = (error,)


class PercentFormatMixedPositionalAndNamed(Message):
    message = "'...' %% ... has mixed positional and named placeholders"


class PercentFormatUnsupportedFormatCharacter(Message):
    message = "'...' %% ... has unsupported format character %r"

    def __init__(self, filename, loc, c):
        Message.__init__(self, filename, loc)
        self.msg_args = (c,)


class PercentFormatPositionalCountMismatch(Message):
    message = "'...' %% ... has %d placeholder(s) but %d substitution(s)"

    def __init__(self, filename, loc, n_placeholders, n_substitutions):
        Message.__init__(self, filename, loc)
        self.msg_args = (n_placeholders, n_substitutions)


class PercentFormatExtraNamedArguments(Message):
    message = "'...' %% ... has unused named argument(s): %s"

    def __init__(self, filename, loc, extra_keywords):
        Message.__init__(self, filename, loc)
        self.msg_args = (extra_keywords,)


class PercentFormatMissingArgument(Message):
    message = "'...' %% ... is missing argument(s) for placeholder(s): %s"

    def __init__(self, filename, loc, missing_arguments):
        Message.__init__(self, filename, loc)
        self.msg_args = (missing_arguments,)


class PercentFormatExpectedMapping(Message):
    message = "'...' %% ... expected mapping but got sequence"


class PercentFormatExpectedSequence(Message):
    message = "'...' %% ... expected sequence but got mapping"


class PercentFormatStarRequiresSequence(Message):
    message = "'...' %% ... `*` specifier requires sequence"



"""
Tests for behaviour related to type annotations.
"""

from sys import version_info

from pyflakes import messages as m
from pyflakes.test.harness import TestCase, skipIf


class TestTypeAnnotations(TestCase):

    def test_typingOverload(self):
        """Allow intentional redefinitions via @typing.overload"""
        self.flakes("""
        import typing
        from typing import overload

        @overload
        def f(s: None) -> None:
            pass

        @overload
        def f(s: int) -> int:
            pass

        def f(s):
            return s

        @typing.overload
        def g(s: None) -> None:
            pass

        @typing.overload
        def g(s: int) -> int:
            pass

        def g(s):
            return s
        """)

    def test_typingExtensionsOverload(self):
        """Allow intentional redefinitions via @typing_extensions.overload"""
        self.flakes("""
        import typing_extensions
        from typing_extensions import overload

        @overload
        def f(s: None) -> None:
            pass

        @overload
        def f(s: int) -> int:
            pass

        def f(s):
            return s

        @typing_extensions.overload
        def g(s: None) -> None:
            pass

        @typing_extensions.overload
        def g(s: int) -> int:
            pass

        def g(s):
            return s
        """)

    def test_typingOverloadAsync(self):
        """Allow intentional redefinitions via @typing.overload (async)"""
        self.flakes("""
        from typing import overload

        @overload
        async def f(s: None) -> None:
            pass

        @overload
        async def f(s: int) -> int:
            pass

        async def f(s):
            return s
        """)

    def test_overload_with_multiple_decorators(self):
        self.flakes("""
            from typing import overload
            dec = lambda f: f

            @dec
            @overload
            def f(x: int) -> int:
                pass

            @dec
            @overload
            def f(x: str) -> str:
                pass

            @dec
            def f(x): return x
       """)

    def test_overload_in_class(self):
        self.flakes("""
        from typing import overload

        class C:
            @overload
            def f(self, x: int) -> int:
                pass

            @overload
            def f(self, x: str) -> str:
                pass

            def f(self, x): return x
        """)

    def test_aliased_import(self):
        """Detect when typing is imported as another name"""
        self.flakes("""
        import typing as t

        @t.overload
        def f(s: None) -> None:
            pass

        @t.overload
        def f(s: int) -> int:
            pass

        def f(s):
            return s
        """)

    def test_not_a_typing_overload(self):
        """regression test for @typing.overload detection bug in 2.1.0"""
        self.flakes("""
            def foo(x):
                return x

            @foo
            def bar():
                pass

            def bar():
                pass
        """, m.RedefinedWhileUnused)

    def test_variable_annotations(self):
        def undefined_names_before_py314(*, n: int):
            return (m.UndefinedName,) * n if version_info < (3, 14) else ()

        self.flakes('''
        name: str
        age: int
        ''')
        self.flakes('''
        name: str = 'Bob'
        age: int = 18
        ''')
        self.flakes('''
        class C:
            name: str
            age: int
        ''')
        self.flakes('''
        class C:
            name: str = 'Bob'
            age: int = 18
        ''')
        self.flakes('''
        def f():
            name: str
            age: int
        ''', m.UnusedAnnotation, m.UnusedAnnotation)
        self.flakes('''
        def f():
            name: str = 'Bob'
            age: int = 18
            foo: not_a_real_type = None
        ''', m.UnusedVariable, m.UnusedVariable, m.UnusedVariable, m.UndefinedName)
        self.flakes('''
        def f():
            name: str
            print(name)
        ''', m.UndefinedName)
        self.flakes('''
        from typing import Any
        def f():
            a: Any
        ''', m.UnusedAnnotation)
        self.flakes('''
        foo: not_a_real_type
        ''', m.UndefinedName)
        self.flakes('''
        foo: not_a_real_type = None
        ''', m.UndefinedName)
        self.flakes('''
        class C:
            foo: not_a_real_type
        ''', m.UndefinedName)
        self.flakes('''
        class C:
            foo: not_a_real_type = None
        ''', m.UndefinedName)
        self.flakes('''
        def f():
            class C:
                foo: not_a_real_type
        ''', m.UndefinedName)
        self.flakes('''
        def f():
            class C:
                foo: not_a_real_type = None
        ''', m.UndefinedName)
        self.flakes('''
        from foo import Bar
        bar: Bar
        ''')
        self.flakes('''
        from foo import Bar
        bar: 'Bar'
        ''')
        self.flakes('''
        import foo
        bar: foo.Bar
        ''')
        self.flakes('''
        import foo
        bar: 'foo.Bar'
        ''')
        self.flakes('''
        from foo import Bar
        def f(bar: Bar): pass
        ''')
        self.flakes('''
        from foo import Bar
        def f(bar: 'Bar'): pass
        ''')
        self.flakes('''
        from foo import Bar
        def f(bar) -> Bar: return bar
        ''')
        self.flakes('''
        from foo import Bar
        def f(bar) -> 'Bar': return bar
        ''')
        self.flakes('''
        bar: 'Bar'
        ''', m.UndefinedName)
        self.flakes('''
        bar: 'foo.Bar'
        ''', m.UndefinedName)
        self.flakes('''
        from foo import Bar
        bar: str
        ''', m.UnusedImport)
        self.flakes('''
        from foo import Bar
        def f(bar: str): pass
        ''', m.UnusedImport)
        self.flakes('''
        def f(a: A) -> A: pass
        class A: pass
        ''', *undefined_names_before_py314(n=2))

        self.flakes('''
        def f(a: 'A') -> 'A': return a
        class A: pass
        ''')
        self.flakes('''
        a: A
        class A: pass
        ''', *undefined_names_before_py314(n=1))
        self.flakes('''
        a: 'A'
        class A: pass
        ''')
        self.flakes('''
        T: object
        def f(t: T): pass
        ''', *undefined_names_before_py314(n=1))
        self.flakes('''
        T: object
        def g(t: 'T'): pass
        ''')
        self.flakes('''
        a: 'A B'
        ''', m.ForwardAnnotationSyntaxError)
        self.flakes('''
        a: 'A; B'
        ''', m.ForwardAnnotationSyntaxError)
        self.flakes('''
        a: '1 + 2'
        ''')
        self.flakes('''
        a: 'a: "A"'
        ''', m.ForwardAnnotationSyntaxError)

    def test_variable_annotation_references_self_name_undefined(self):
        self.flakes("""
        x: int = x
        """, m.UndefinedName)

    def test_TypeAlias_annotations(self):
        self.flakes("""
        from typing_extensions import TypeAlias
        from foo import Bar

        bar: TypeAlias = Bar
        """)
        self.flakes("""
        from typing_extensions import TypeAlias
        from foo import Bar

        bar: TypeAlias = 'Bar'
        """)
        self.flakes("""
        from typing_extensions import TypeAlias
        from foo import Bar

        class A:
            bar: TypeAlias = Bar
        """)
        self.flakes("""
        from typing_extensions import TypeAlias
        from foo import Bar

        class A:
            bar: TypeAlias = 'Bar'
        """)
        self.flakes("""
        from typing_extensions import TypeAlias

        bar: TypeAlias
        """)
        self.flakes("""
        from typing_extensions import TypeAlias
        from foo import Bar

        bar: TypeAlias
        """, m.UnusedImport)

    def test_annotating_an_import(self):
        self.flakes('''
            from a import b, c
            b: c
            print(b)
        ''')

    def test_unused_annotation(self):
        # Unused annotations are fine in module and class scope
        self.flakes('''
        x: int
        class Cls:
            y: int
        ''')
        self.flakes('''
        def f():
            x: int
        ''', m.UnusedAnnotation)
        # This should only print one UnusedVariable message
        self.flakes('''
        def f():
            x: int
            x = 3
        ''', m.UnusedVariable)

    def test_unused_annotation_in_outer_scope_reassigned_in_local_scope(self):
        self.flakes('''
        x: int
        x.__dict__
        def f(): x = 1
        ''', m.UndefinedName, m.UnusedVariable)

    def test_unassigned_annotation_is_undefined(self):
        self.flakes('''
        name: str
        print(name)
        ''', m.UndefinedName)

    def test_annotated_async_def(self):
        self.flakes('''
        class c: pass
        async def func(c: c) -> None: pass
        ''')

    def test_postponed_annotations(self):
        self.flakes('''
        from __future__ import annotations
        def f(a: A) -> A: pass
        class A:
            b: B
        class B: pass
        ''')

        self.flakes('''
        from __future__ import annotations
        def f(a: A) -> A: pass
        class A:
            b: Undefined
        class B: pass
        ''', m.UndefinedName)

        self.flakes('''
        from __future__ import annotations
        T: object
        def f(t: T): pass
        def g(t: 'T'): pass
        ''')

    def test_annotations_do_not_define_names_with_future_annotations(self):
        self.flakes('''
            from __future__ import annotations
            def f():
                x: str
                print(x)
        ''', m.UndefinedName)

    @skipIf(version_info < (3, 14), 'new in Python 3.14')
    def test_postponed_annotations_py314(self):
        self.flakes('''
            def f(x: C) -> None: pass
            class C: pass
        ''')

    def test_type_annotation_clobbers_all(self):
        self.flakes('''\
        from typing import TYPE_CHECKING, List

        from y import z

        if not TYPE_CHECKING:
            __all__ = ("z",)
        else:
            __all__: List[str]
        ''')

    def test_return_annotation_is_class_scope_variable(self):
        self.flakes("""
        from typing import TypeVar
        class Test:
            Y = TypeVar('Y')

            def t(self, x: Y) -> Y:
                return x
        """)

    def test_return_annotation_is_function_body_variable(self):
        self.flakes("""
        class Test:
            def t(self) -> Y:
                Y = 2
                return Y
        """, m.UndefinedName)

    def test_positional_only_argument_annotations(self):
        self.flakes("""
        from x import C

        def f(c: C, /): ...
        """)

    def test_partially_quoted_type_annotation(self):
        self.flakes("""
        from queue import Queue
        from typing import Optional

        def f() -> Optional['Queue[str]']:
            return None
        """)

    def test_partially_quoted_type_assignment(self):
        self.flakes("""
        from queue import Queue
        from typing import Optional

        MaybeQueue = Optional['Queue[str]']
        """)

    def test_nested_partially_quoted_type_assignment(self):
        self.flakes("""
        from queue import Queue
        from typing import Callable

        Func = Callable[['Queue[str]'], None]
        """)

    def test_quoted_type_cast(self):
        self.flakes("""
        from typing import cast, Optional

        maybe_int = cast('Optional[int]', 42)
        """)

    def test_type_cast_literal_str_to_str(self):
        # Checks that our handling of quoted type annotations in the first
        # argument to `cast` doesn't cause issues when (only) the _second_
        # argument is a literal str which looks a bit like a type annotation.
        self.flakes("""
        from typing import cast

        a_string = cast(str, 'Optional[int]')
        """)

    def test_quoted_type_cast_renamed_import(self):
        self.flakes("""
        from typing import cast as tsac, Optional as Maybe

        maybe_int = tsac('Maybe[int]', 42)
        """)

    def test_quoted_TypeVar_constraints(self):
        self.flakes("""
        from typing import TypeVar, Optional

        T = TypeVar('T', 'str', 'Optional[int]', bytes)
        """)

    def test_quoted_TypeVar_bound(self):
        self.flakes("""
        from typing import TypeVar, Optional, List

        T = TypeVar('T', bound='Optional[int]')
        S = TypeVar('S', int, bound='List[int]')
        """)

    def test_literal_type_typing(self):
        self.flakes("""
        from typing import Literal

        def f(x: Literal['some string']) -> None:
            return None
        """)

    def test_literal_type_typing_extensions(self):
        self.flakes("""
        from typing_extensions import Literal

        def f(x: Literal['some string']) -> None:
            return None
        """)

    def test_annotated_type_typing_missing_forward_type(self):
        self.flakes("""
        from typing import Annotated

        def f(x: Annotated['integer']) -> None:
            return None
        """, m.UndefinedName)

    def test_annotated_type_typing_missing_forward_type_multiple_args(self):
        self.flakes("""
        from typing import Annotated

        def f(x: Annotated['integer', 1]) -> None:
            return None
        """, m.UndefinedName)

    def test_annotated_type_typing_with_string_args(self):
        self.flakes("""
        from typing import Annotated

        def f(x: Annotated[int, '> 0']) -> None:
            return None
        """)

    def test_annotated_type_typing_with_string_args_in_union(self):
        self.flakes("""
        from typing import Annotated, Union

        def f(x: Union[Annotated['int', '>0'], 'integer']) -> None:
            return None
        """, m.UndefinedName)

    def test_literal_type_some_other_module(self):
        """err on the side of false-negatives for types named Literal"""
        self.flakes("""
        from my_module import compat
        from my_module.compat import Literal

        def f(x: compat.Literal['some string']) -> None:
            return None
        def g(x: Literal['some string']) -> None:
            return None
        """)

    def test_literal_union_type_typing(self):
        self.flakes("""
        from typing import Literal

        def f(x: Literal['some string', 'foo bar']) -> None:
            return None
        """)

    def test_deferred_twice_annotation(self):
        self.flakes("""
            from queue import Queue
            from typing import Optional


            def f() -> "Optional['Queue[str]']":
                return None
        """)

    def test_partial_string_annotations_with_future_annotations(self):
        self.flakes("""
            from __future__ import annotations

            from queue import Queue
            from typing import Optional


            def f() -> Optional['Queue[str]']:
                return None
        """)

    def test_forward_annotations_for_classes_in_scope(self):
        # see #749
        self.flakes("""
        from typing import Optional

        def f():
            class C:
                a: "D"
                b: Optional["D"]
                c: "Optional[D]"

            class D: pass
        """)

    def test_idomiatic_typing_guards(self):
        # typing.TYPE_CHECKING: python3.5.3+
        self.flakes("""
            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                from t import T

            def f() -> T:
                pass
        """)
        # False: the old, more-compatible approach
        self.flakes("""
            if False:
                from t import T

            def f() -> T:
                pass
        """)
        # some choose to assign a constant and do it that way
        self.flakes("""
            MYPY = False

            if MYPY:
                from t import T

            def f() -> T:
                pass
        """)

    def test_typing_guard_for_protocol(self):
        self.flakes("""
            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                from typing import Protocol
            else:
                Protocol = object

            class C(Protocol):
                def f() -> int:
                    pass
        """)

    def test_typednames_correct_forward_ref(self):
        self.flakes("""
            from typing import TypedDict, List, NamedTuple

            List[TypedDict("x", {})]
            List[TypedDict("x", x=int)]
            List[NamedTuple("a", a=int)]
            List[NamedTuple("a", [("a", int)])]
        """)
        self.flakes("""
            from typing import TypedDict, List, NamedTuple, TypeVar

            List[TypedDict("x", {"x": "Y"})]
            List[TypedDict("x", x="Y")]
            List[NamedTuple("a", [("a", "Y")])]
            List[NamedTuple("a", a="Y")]
            List[TypedDict("x", {"x": List["a"]})]
            List[TypeVar("A", bound="C")]
            List[TypeVar("A", List["C"])]
        """, *[m.UndefinedName]*7)
        self.flakes("""
            from typing import NamedTuple, TypeVar, cast
            from t import A, B, C, D, E

            NamedTuple("A", [("a", A["C"])])
            TypeVar("A", bound=A["B"])
            TypeVar("A", A["D"])
            cast(A["E"], [])
        """)

    def test_namedtypes_classes(self):
        self.flakes("""
            from typing import TypedDict, NamedTuple
            class X(TypedDict):
                y: TypedDict("z", {"zz":int})

            class Y(NamedTuple):
                y: NamedTuple("v", [("vv", int)])
        """)

    @skipIf(version_info < (3, 11), 'new in Python 3.11')
    def test_variadic_generics(self):
        self.flakes("""
            from typing import Generic
            from typing import TypeVarTuple

            Ts = TypeVarTuple('Ts')

            class Shape(Generic[*Ts]): pass

            def f(*args: *Ts) -> None: ...

            def g(x: Shape[*Ts]) -> Shape[*Ts]: ...
        """)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_statements(self):
        self.flakes("""
            type ListOrSet[T] = list[T] | set[T]

            def f(x: ListOrSet[str]) -> None: ...

            type RecursiveType = int | list[RecursiveType]

            type ForwardRef = int | C

            type ForwardRefInBounds[T: C] = T

            class C: pass
        """)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_functions(self):
        self.flakes("""
            def f[T](t: T) -> T: return t

            async def g[T](t: T) -> T: return t

            def with_forward_ref[T: C](t: T) -> T: return t

            def can_access_inside[T](t: T) -> T:
                print(T)
                return t

            class C: pass
        """)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_do_not_escape_function_scopes(self):
        self.flakes("""
            from x import g

            @g(T)  # not accessible in decorators
            def f[T](t: T) -> T: return t

            T  # not accessible afterwards
        """, m.UndefinedName, m.UndefinedName)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_classes(self):
        self.flakes("""
            class C[T](list[T]): pass

            class UsesForward[T: Forward](list[T]): pass

            class Forward: pass

            class WithinBody[T](list[T]):
                t = T
        """)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_do_not_escape_class_scopes(self):
        self.flakes("""
            from x import g

            @g(T)  # not accessible in decorators
            class C[T](list[T]): pass

            T  # not accessible afterwards
        """, m.UndefinedName, m.UndefinedName)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_TypeVarTuple(self):
        self.flakes("""
        def f[*T](*args: *T) -> None: ...
        """)

    @skipIf(version_info < (3, 12), 'new in Python 3.12')
    def test_type_parameters_ParamSpec(self):
        self.flakes("""
        from typing import Callable

        def f[R, **P](f: Callable[P, R]) -> Callable[P, R]:
            def g(*args: P.args, **kwargs: P.kwargs) -> R:
                return f(*args, **kwargs)
            return g
        """)

    @skipIf(version_info < (3, 13), 'new in Python 3.13')
    def test_type_parameter_defaults(self):
        self.flakes("""
        def f[T = int](u: T) -> T:
            return u
        """)



"""
This module provides the configuration interface for Drace —
both as a command-line utility and an interactive USSD-like
menu. It supports setting, resetting, and retrieving
persistent user-defined options for Drace's behavior.

Key Features:
- Persistent config management via a JSON file
- CLI commands to view, update, or reset individual or all
  config options
- USSD-style interactive interface for ease of use
- Supports type-aware casting, list manipulation
  (add/remove), and safe validation

Functions:
- config_cmd: Entry point for handling config commands from
  CLI
- sanitize_args: Parses and normalizes CLI arguments
- interactive: Interactive USSD-like interface for setting
  config options
- _handle_list: Handles list config operations (+ append, -
  remove)
- list_items, choose, transmit, etc. assist in formatting and
  interaction

Classes:
- Config: Manages loading, saving, setting, and resetting of
          config values

Notes:
- The term 'hapana' (Shona for "nothing") is used internally
  to represent a placeholder value in scenarios where no
  explicit value is given.

This module is designed to be user-friendly, flexible, and
forgiving — in line with Drace's pragmatic philosophy.
"""
# ========================= STANDARDS =======================
from typing import NoReturn
import json
import sys
import os
import re

# ======================= THIRD-PARTIES =====================
from tuikit.listools import list_items, choose, pick_from
from tuikit.logictools import any_eq, any_in
from tuikit.textools import strip_ansi

# ========================== LOCALS =========================
from .utils import APP_COLOR, YELLOW, GOOD, BAD, CURSOR, CMDS
from .utils import DEFAULTS_PATH, transmit, center, color
from .docs import config as doc
from .utils import underline
from . import help_menu


__doc__  = doc  # override for calls
DEFAULTS = {
    "mode":          [str, "lint"],
    "line_len":      [int, 79],
    "max_fn_steps":  [int, 6],
    "max_coupling":  [int, 3],
    "wrap":          [bool, True],
    "color":         [bool, True],
    "score":         [bool, True],
    "only_rules":    [list, None],
    "ignored_rules": [list, None],
    "ignored_files": [list, None],
    "delay":         [float, 0.01]
}


class Config:
    """
    Handles persistent user configuration for Drace.

    Loads, sets, and resets user-defined settings. Values are
    saved to a JSON file (`DEFAULTS_PATH`), and changes are
    applied instantly.

    This class is tightly integrated with the CLI and
    USSD-style interface.

    Attributes:
        path (str): Path to the config file.
        defaults (dict): The current active defaults.
    """
    def __init__(self):
        """Initializes the config system by loading existing or default values."""
        self.path     = DEFAULTS_PATH
        self.defaults = self._load()

    def _load(self) -> dict:
        """
        Loads config from file. If it doesn't exist, calls
        `reset()` to generate default values.

        Returns:
            dict: The active configurations
        """
        if os.path.exists(self.path):
            with open(self.path) as f: return json.load(f)

        self.reset()
        return self.defaults

    def _save(self) -> None:
        """Persists the current defaults to json."""
        with open(self.path, 'w') as f:
            json.dump(self.defaults, f)

    def set(self, key: str, value: int | float | bool | list):
        """
        Sets a config value and saves to json.

        Args:
            key (str): The config key
            value (int|float|bool|list): The new value to
                                         assign
        """
        self.defaults[key] = value
        self._save()

    def get(self, key: str, fallback=None
           ) -> int | float | bool | list | None:
        """
        Retrieves a config value.

        Args:
            key (str): The config key.
            fallback: Optional value to return if key doesn't
                      exist

        Returns:
            Any: The config value or fallback.
        """
        return self.defaults.get(key, fallback)

    def reset(self, target: str = "hapana"):
        """
        Resets one or all config values to their defaults.

        Args:
            target (str): The key to reset, or "hapana"/"all"
                          to reset everything
        """
        if target and not any_eq("hapana", "all", eq=target):
            if target not in DEFAULTS:
                transmit(f"unknown reset target: {target}\n")
                return
            self.defaults[target] = DEFAULTS[target][1]
        else:
            defaults = {}
            for k, v in DEFAULTS.items(): defaults[k] = v[1]
            self.defaults = defaults
        self._save()


def sanitize_args(args) -> list[str] | NoReturn:
    """
    Normalizes and interprets CLI arguments for `drace
    config`.

    Handles flexible separators (`=`, `:`, `+`, `-`) and
    reconstructs input into a standard [key, value(s)]
    format. Also simulates modifier flags (`+`, `-`) by
    appending them to `sys.argv` for later checks.

    Special handling for:
      - Single argument lookups (e.g. `drace config wrap`)
      - Key-only commands like `reset`, `show`, `list`, and
        `help`
      - Embedded separators (e.g. `line_len=100`,
        `ignored_rules+Z100`)
      - Gracefully handles unknown keys with user feedback

    'hapana' is used as a placeholder value when only a key
    is passed (e.g. `drace config line_len`) to differentiate
    it from `None` or other falsy values — since those might
    be valid defaults. This allows `config_cmd()` to detect
    intent (like showing a value) without conflict.

    Args:
        args (list): Raw CLI arguments passed to the config
                     command

    Returns:
        list: A cleaned list of [key, value(s)] or control
              command
    """
    sep = ["=", ":", "+", "-"]

    if args:
        if len(args) > 2:
            clean = [args[0]]
            if any_in(sep, eq=args[1]):
                if len(args) > 3: clean.append(args[2:])
                else: clean.append(args[2])
            else: clean.append(args[1:])
            args = clean
        elif len(args) == 2:
            clean = []
            for arg in args:
                # allow for an arbitrary number of separators
                for s in sep:
                    if s in arg:
                        sys.argv.append(s)
                        s  *= arg.count(s)
                        arg = arg.replace(s, "")
                clean.append(arg)
            args = clean
        else:
            if any_in(sep, eq=args[0]):
                # allow for an arbitrary number of separators
                sep[0] *= max(args[0].count("="), 1)
                sep[1] *= max(args[0].count(":"), 1)
                pattern = "|".join(map(re.escape, sep)) \
                        + r"|\s+"
                args    = re.split(pattern, args[0])
            else:
                extras = ["reset", "show", "list", "help"]
                args   = args[0]
                if args not in config.defaults:
                    if args not in extras:
                        transmit(f"unknown key: {args}\n",
                                  BAD)
                        sys.exit(1)
                args = [args, "hapana"]

        if isinstance(args[1], str):
            args[1] = args[1].split()
            if len(args[1]) == 1: args[1] = args[1][0]

    return args


def interactive() -> list[str] | NoReturn:
    """
    Launches an interactive USSD-like menu for configuring
    Drace defaults.

    Prompts the user to select a config option and input a
    new value.

    Displays current value and hints. Gracefully handle
    exits (via Ctrl+C, Ctrl+D, or explicit "Back" choice).

    Returns a list containing the selected config key and its
    new value.
    """
    head = center("《 DRACE CONFIG 》", "=", GOOD, APP_COLOR)
    print(f"\n{head}")

    cmd = color("drace config reset", YELLOW)
    transmit(f"use {cmd} to restore defaults\n")

    # list keys USSD-style and get choice
    list_items(DEFAULTS, guide="Choose option to change")
    choice = choose(DEFAULTS, getch=True, src=interactive,
             hue=APP_COLOR)

    if choice is None: print(); sys.exit(0)

    # retrieve value and present to user
    option   = pick_from(DEFAULTS.keys(), choice)
    value    = config.get(option)
    key, val = color(option, GOOD), color(value, YELLOW)
    message  = f"{key} ::: {val} ::: New value:"
    transmit(message)

    try: new = input(CURSOR).strip()
    except (KeyboardInterrupt, EOFError) as e:
        if isinstance(e, EOFError): print()
        sys.exit(1)
    finally: underline(hue=APP_COLOR, alone=True)

    return [option, new]


def _handle_list(key: str, values: tuple[str | list]
                ) -> list[str] | None:
    """
    Handles list-type config value manipulation (set, add,
    remove).

    Determines whether to:
      - replace the entire list,
      - append new values (if '+' present in sys.argv),
      - remove specified values (if '-' present in sys.argv)

    Args:
        key (str): The config key being modified
        values (list): The new value(s) passed from CLI input

    Returns:
        list | None: The updated list for the given key, or
                     None if resulting list is empty

    Notes:
        - Values are extended or removed based on presence of
          '+' or '-' in sys.argv
        - If the final list contains only one empty string,
          it's interpreted as None
    """
    if isinstance(values[0], str): value = list(values)
    else: value = values[0]

    array = config.get(key) or []

    if "+" in " ".join(sys.argv):  # append
        array.extend(value)
    elif "-" in " ".join(sys.argv):  # remove
        for val in values:
            try: array.remove(val)
            except ValueError: pass
    else: array = value  # override

    if len(array) == 1 and not array[0]: array = None
    if array is not None:
        for i, value in enumerate(array):
            if os.sep in value:
                value = value.split(os.sep)[-1]
            array[i] = value

    return array


def config_cmd(args: list[str] | list) -> None:
    """
    Handles configuration commands for Drace via C9LI or
    interactive prompt.

    Supports viewing, modifying, and resetting user-defined
    configuration options. The function also handles
    displaying help information, listing all current config
    values, and updating individual keys.

    Args:
        args (list[str] | list): Command-line arguments
            passed to `drace config`. If empty or None,
            launches the interactive config menu.

    Supported commands:
        - <key> <value>         Set the value of a config key
        - <key> = <value>       Alternative way to set a
                                value
        - <key> + <value>       Add to list-type config
                                values
        - <key> - <value>       Remove from list-type config
                                values
        - list                  Show all current config
                                values
        - show <key>            Show value of a specific
                                config key
        - reset <key|all>       Reset a key or all config
                                values to defaults
        - help                  how help menu

    Notes:
        - Handles type conversion based on the expected type
          for each key.
        - Allows multiple input styles and separators (e.g.,
          "key ::: value").
        - List-type configs can be modified incrementally.
        - Ignores case for boolean values ("yes", "on", etc)
    """
    def list_all(args: list[str]) -> bool:
        for cmd in ["list"] + list(config.defaults.keys()):
            if args[0] == cmd and args[1] == "hapana":
                return True
        return False

    if not args: args = interactive()

    args = sanitize_args(args)

    if list_all(args):
        transmit("config:")
        for k, v in config.defaults.items():
            s = ":" * (25 - len(k))
            k = color(k, GOOD)
            v = color(v, YELLOW)
            if args and args[0] != "list":
                if args[0] == strip_ansi(k):
                    transmit(f"{k} ::: {v}\n", _list=True)
                    return
            else: transmit(f"{k} {s} {v}", _list=True)
        print(); return

    if args[0] == "help": help_menu.main("config")
    if args[0] == "reset": config.reset(args[1]); return

    if args[0] == "show":
        try:
            key   = color(args[1], GOOD)
            value = config.get(args[1])
            transmit(f"{key} ::: {color(value, YELLOW)}\n")
        except KeyError:
            transmit(f"Unknown key: {args[1]}\n", BAD)
        return

    key, *values = args

    key = key.strip()
    if key not in DEFAULTS:
        transmit(f"unknown key: {key}\n", BAD)
        return

    _type    = DEFAULTS[key][0]
    truthy   = ["1", "true", "yeah", "yes", "y", "on", "ok"]
    exp_list = ""
    try:
        if _type == bool: value = values[0].lower() in truthy
        elif _type == list: value = _handle_list(key, values)
        else:
            if _type == str and values[0] not in CMDS:
                exp_list += ", ".join(CMDS[:-1])
                exp_list += ", or " + CMDS[-1]
                raise ValueError
            value = _type(values[0])
    except (IndexError, ValueError, TypeError):
        transmit(f"Invalid value for {key}. Expected "
               + f"{exp_list or _type.__name__}\n", BAD)
        return

    config.set(key, value)


config = Config()



import builtins
import ast

from drace.types import Context, Dict


def _is_not_synthetic_name(name: str) -> bool:
    if any(name.startswith(p) for p in ["_V", "_A"]):
        try: int(name[2]); return False
        except ValueError: pass

    return True


def iter_used_names(node: ast.AST) -> set[str]:
    used: set[str] = set()

    def visit(n: ast.AST):
        # Stop at nested scopes
        if n is not node and isinstance(n,
            (ast.FunctionDef, ast.AsyncFunctionDef,
            ast.ClassDef, ast.Lambda)): return

        if isinstance(n, ast.Name) and isinstance(n.ctx,
                ast.Load):
            if _is_not_synthetic_name(n.id): used.add(n.id)

        for child in ast.iter_child_nodes(n): visit(child)

    visit(node)
    return used


def get_enclosing_locals(node: ast.AST) -> set[str]:
    """
    Collect names visible from enclosing *non-module* scopes.
    Module scope is handled explicitly elsewhere.
    """
    visible = set()
    cur     = node.parent

    while cur and not isinstance(cur, ast.Module):
        # -------- Function / AsyncFunction scope --------
        if isinstance(cur, (ast.FunctionDef,
                ast.AsyncFunctionDef)):
            # Parameters
            for arg in cur.args.args + cur.args.kwonlyargs:
                visible.add(arg.arg)
            if cur.args.vararg:
                visible.add(cur.args.vararg.arg)
            if cur.args.kwarg:
                visible.add(cur.args.kwarg.arg)

            # Local assignments
            for sub in cur.body:
                if isinstance(sub, (ast.Assign,
                        ast.AnnAssign)):
                    targets = sub.targets if hasattr(sub,
                              "targets") else [sub.target]
                    for t in targets:
                        visible |= extract_names(t)

                elif isinstance(sub, (ast.For,
                        ast.AsyncFor)):
                    visible |= extract_names(sub.target)

                elif isinstance(sub, ast.ExceptHandler) \
                    and sub.name: visible.add(sub.name)

                elif isinstance(sub, (ast.FunctionDef,
                        ast.AsyncFunctionDef, ast.ClassDef)):
                    visible.add(sub.name)

        # -------- Class scope --------
        elif isinstance(cur, ast.ClassDef):
            # Only the class name is visible to inner scopes
            visible.add(cur.name)

        # Move upward
        cur = cur.parent

    return visible


def extract_names(node):
    """
    Recursively extract name identifiers from ast targets
    (for assignment, tuple unpacking, etc).
    """
    names: set[str] = set()
    if isinstance(node, ast.Name): names.add(node.id)
    elif isinstance(node, (ast.Tuple, ast.List)):
        for elt in node.elts: names |= extract_names(elt)
    return names


def get_targets(node: ast.AST) -> list:
    return node.targets \
        if hasattr(node, "targets") \
      else [node.target]


def check_z227(context: Context) -> list[Dict]:
    """
    Z227: Detect hidden dependencies in functions and nested
          functions.

    Hidden dependency: a name used in a function that is not:
      - a parameter or local variable (or nested def/class),
      - an imported name,
      - a built-in,
      - a module-level constant (ALL_CAPS assigned at top
        level).

    Globals (non-constant) and other external names will be
    flagged.
    """
    tree    = context["tree"]
    file    = context["file"]
    results = []

    builtin_names: set[str] = set(dir(builtins))

    # 1. Collect imported names (module‑level)
    imported_names: set[str] = set()
    for node in tree.body:
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            for alias in node.names:
                imported_names.add(alias.asname or
                    alias.name.split(".")[0])

    # 2. Collect module‑level variable names → identify
    #    constants
    module_globals:   set[str] = set()
    module_constants: set[str] = set()
    for node in tree.body:
        if isinstance(node, (ast.Assign, ast.AnnAssign)):
            targets = get_targets(node)
            for t in targets:
                if isinstance(t, ast.Name):
                    module_globals.add(t.id)
                    if t.id.isupper():
                        module_constants.add(t.id)
                elif isinstance(t, (ast.Tuple, ast.List)):
                    for elt in t.elts:
                        if isinstance(elt, ast.Name):
                            module_globals.add(elt.id)
                            if elt.id.isupper():
                                module_constants.add(elt.id)
        elif isinstance(node, (ast.FunctionDef,
                ast.AsyncFunctionDef, ast.ClassDef)):
            module_globals.add(node.name)

    # 3. Walk through all functions and check for hidden deps
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef,
                ast.AsyncFunctionDef, ast.Lambda)):
            if isinstance(node.parent, ast.Module): continue
            # Collect used names inside this function
            used_names = iter_used_names(node)

            # Collect locals: arguments, names assigned
            # within the function body, names of nested
            # defs/classes
            local_vars: set[str] = set()
            args = node.args
            for arg in list(args.args) + list(
                args.kwonlyargs): local_vars.add(arg.arg)
            if args.vararg: local_vars.add(args.vararg.arg)
            if args.kwarg: local_vars.add(args.kwarg.arg)
            for sub in ast.iter_child_nodes(node):
                if isinstance(sub, (ast.FunctionDef,
                        ast.AsyncFunctionDef, ast.ClassDef,
                        ast.Lambda)):
                    local_vars.add(sub.name if hasattr(sub,
                        "name") else "<lambda>")
            # Also treat names assigned inside as locals
            for sub in ast.walk(node):
                if isinstance(sub, (ast.Assign,
                        ast.AnnAssign)):
                    targets = get_targets(sub)
                    for t in targets:
                        local_vars |= extract_names(t)
                elif isinstance(sub, (ast.For,
                        ast.comprehension)):
                    target = getattr(sub, "target", None)
                    if target is not None:
                        local_vars |= extract_names(target)
                elif isinstance(sub, ast.ExceptHandler
                    ) and sub.name: local_vars.add(sub.name)

            # Determine visible names from enclosing
            # (non-module) scopes
            visible_from_enclosing = get_enclosing_locals(
                                     node)

            # Now compute hidden dependencies
            hidden = used_names - local_vars \
                   - visible_from_enclosing - builtin_names \
                   - imported_names - module_constants \
                   - module_globals

            if hidden:
                results.append({
                    'file': file,
                    'line': node.lineno,
                    'col': 1,
                    'code': 'Z227',
                    'msg': "function uses hidden "
                          f"dependencies: {sorted(hidden)}"
                })

    return results



from typing import TypedDict
import ast


class Context(TypedDict):
    lines: list[str]
    tree:  ast.Module
    file:  str


class Dict(TypedDict):
    file: str
    line: int
    col:  int
    code: str
    msg:  str



# ========================= STANDARDS =======================
import sys

# ======================= THIRD-PARTIES =====================
from tuikit.logictools import any_in

# ========================== LOCALS =========================
from .utils import GOOD, APP_COLOR, center, wrap_text, color
from .utils import underline


def main(to: str | None = None):
    def heading():
        nonlocal header
        print(f"\n{header}\n")

    head   = f" {to.upper()} " if to is not None else " "
    header = center(f"《 DRACE{head}HELP 》", "=", APP_COLOR,
             GOOD)

    if to:
        heading()
        if to == "config": config()
    else:
        desc = drace(heading)
        if desc is not None: return desc

    underline(hue=GOOD, alone=True)
    sys.exit(0)


def drace(heading):
    if not any_in("-h", "-help", eq=sys.argv): return ""
    heading()
    text = f"""Usage: 
    drace [command] [options] <path>

Commands:

1. lint 
{wrap_text("Lint Python files and show potential issues, suggestions, or design improvements.", 3, 3)}
   Example: drace lint format src/

2. format  
{wrap_text("Format Python files using Drace’s code styling rules.", 3, 3)}
   Example: drace myfile.py --diff

3. score  
{wrap_text("Calculate a linting score based on rules triggered.", 3, 3)}
   Example: drace score .

4. config
{wrap_text("Configure defaults either interactively or via command line.", 3, 3)}
{wrap_text("Example: drace config line_len 100", 3, 3)}
{wrap_text("Example: drace config (opens interactive mode)", 3, 3)}

Common Options:

  --score         {wrap_text("Toggle score after linting or formatting; relative to default set", 18, inline=True, order="                  ")}
  --color         {wrap_text("Toggle colorized output; relative to default set", 18, inline=True, order="                  ")}
  --diff          {wrap_text("Show diff instead of modifying files (used with format)", 18, inline=True, order="                  ")}
  -h, --help      {wrap_text("Show this help message and exit", 18, inline=True, order="                  ")}

Config Usage:
    drace config help

Notes:
{wrap_text("• Cross-file analysis, resilience to syntax errors, and rule flexibility are part of Drace’s core goals.", 2)}
{wrap_text("• Defaults are user-friendly and can be fine-tuned in seconds.", 2)}

Website / Docs: Coming soon."""
    print(text)


def config():
    # ============ SECTION 1: CHANGING DEFAULTS =============
    print(color("Changing defaults", "", "", True, True))
    
    # subsection 1: format
    print(color("\n  1. Formats:", GOOD))
    print("   • drace config key value")
    print("   • drace config key=value")
    print("   • drace config key = value")
    print("   • drace config key:value")
    print("   • drace config key : value")
    print("   • drace config key + value (appends to list)")
    print("   • drace config key - value (remove from list)")
    
    # subsection 2: examples
    print(color("\n  2. Examples:", GOOD))
    print("   • drace config line_len 59")
    print("   • drace config color :: on")
    print("   • kityy config score====yes")
    print("   • drace config ignored_rules + Z100 Z200...\n")

    # ============ SECTION 2: RESETTING DEFAULTS ============
    print(color("Resetting defaults", "", "", True, True))

    # subsection 1: format
    print(color("\n  1. Formats:", GOOD))
    print("   • drace config reset      Restore defaults")
    print("   • drace config reset all  Same as above")
    print("   • drace config reset key  Reset one key")

    # subsection 2: examples
    print(color("\n  2. Examples:", GOOD))
    print("   • drace config reset")
    print("   • drace config reset all")
    print("   • drace config reset wrap\n")

    # ============= SECTION 3: VIEWING DEFAULTS =============
    print(color("Viewing defaults", "", "", True, True))
    
    # subsection 1: format
    print(color("\n  1. Formats:", GOOD))
    print("   • drace config list       List all defaults")
    print("   • drace config key        Show one default")
    print("   • drace config show key   Same as above")
    
    # subsection 2: examples
    print(color("\n  2. Examples:", GOOD))
    print("   • drace config list")
    print("   • drace config line_len")
    print("   • drace config show line_len\n")

    # ============= SECTION 4: INTERACTIVE MODE =============
    print(color("Interactive mode", "", "", True, True))
    
    # subsection 1: format
    print(color("\n  1. Format:", GOOD))
    print("   • drace config")
    
    # subsection 2: info
    print(color("\n  2. Info:", GOOD))
    print(wrap_text("• Opens a USSD-style interactive menu "
        +"to set defaults", 5, 3))
    print(wrap_text("• The setting rules in <Changing "
        +"defaults> apply here, except you don't have to "
        +"include the key:\n", 5, 3))
    print(wrap_text("   >>> + Z100 Z200  # if you had chosen"
        +" ignored_rules and wanted to append more rules\n", 
        24, 3))

    # ============= SECTION 4: INTERACTIVE MODE =============
    print(color("Notes:", "", "", True, True))
    
    print(wrap_text("• if '-' or '+' is not found when "
        +"setting a list default it will be overriden", 5,
        3))



#!/usr/bin/env python3
"""
This module defines the command-line interface for Drace — a
pragmatic linter and formatter for Python code.  

It handles argument parsing and dispatches execution to the
appropriate command module based on user input.

Commands:
- `drace format <path>`: Format code and optionally show a
   diff
- `drace lint <path>`: Lint Python files and display
   suggestions or warnings
- `drace score <path>`: Compute and display a score based on
   linting results
- `drace config [args]`: Configure default settings
  interactively or via command arguments

Example usage:
    drace lint project/
    drace format script.py --diff
    drace config line_len 100
"""
# ========================= STANDARDS =======================
from typing import NoReturn, Callable
from pathlib import Path
import argparse
import sys
import os

# ========================== LOCALS =========================
from .constants import LINE_LEN, MODE, SCORE, CMDS, override
from .reporters import linting, formatting
from .help_menu import main as drace
from .native_lang import translate
from .config import config_cmd
from .docs import cli
from . import utils


__doc__ = cli


def parse_args(argv: list[str]) -> argparse.Namespace:
    """
    Parse command-line arguments for the Drace CLI

    Supports four subcommands:
    - format: Format Python files with optional diff and
      score display
    - lint: Lint Python files and optionally display a score
    - score: Show the code quality score for a given path
    - config: View or modify Drace configuration (interactive
      or direct)

    Args:
        argv (list): List of command-line arguments
                     (excluding program name)

    Returns:
        argparse.Namespace: Parsed arguments with attributes
                            based on subcommand
    """
    p   = argparse.ArgumentParser(description=drace())
    sub = p.add_subparsers(dest="cmd")

    # formatter
    fmt = sub.add_parser("format")
    fmt.add_argument("path")
    fmt.add_argument("--diff", action="store_true")
    fmt.add_argument("--score", nargs="?", default=SCORE)
    fmt.add_argument("--color", action="store_true")

    # linter
    lint = sub.add_parser("lint")
    lint.add_argument("path")
    lint.add_argument("--score", nargs="?", default=SCORE)
    lint.add_argument("--color", action="store_true")

    # scoring
    score = sub.add_parser("score")
    score.add_argument("path")
    score.add_argument("--color", action="store_true")

    # config
    config = sub.add_parser("config")
    config.add_argument("args", nargs="*")

    return p.parse_args(argv)


def main() -> None | NoReturn:
    """
    Entry point for the Drace CLI

    Parses command-line arguments and dispatches to the 
    appropriate handler:
    - format: Formats the given file or directory
    - lint: Lints one or more Python files
    - score: Displays lint score for the given path
    - config: Opens or modifies configuration
    """
    def workflow(run: Callable) -> int:
        nonlocal exit_code
        files = utils.discover_code_files(path)

        try: score = args.score
        except AttributeError: score = False

        for i, file in enumerate(files):
            file = str(file)
            file, discard = translate(file)
            done = i == len(files) - 1
            try: exc = run(file, score, i == 0, done)
            except KeyboardInterrupt:
                utils.transmit("user aborted\n", utils.BAD)
                sys.exit(1)
            if not exit_code: exit_code = exc
            if discard: os.remove(file)
        return exit_code

    
    try:
        if sys.argv[1] not in CMDS:
            sys.argv.insert(1, MODE)
    except IndexError: sys.argv.extend([MODE, "."])

    args = parse_args(sys.argv[1:])
    try: path = Path(args.path)
    except AttributeError: pass

    args      = override(args)
    exit_code = 0
    
    if args.cmd == "format":
        formatting.format_cmd(path, diff=args.diff)
    elif args.cmd in ["lint", "score"]:
        exit_code = workflow(linting.lint_cmd)
    elif args.cmd == "config":
        config_cmd(args.args)

    sys.exit(exit_code)


if __name__ == "__main__": main()



from drace.constants import KEYWORDS, LINE_LEN
from drace.types import Context, Dict


def check_z200(context: Context) -> list[Dict]:
    """
    Z200: Encourage compact control blocks.

    Flags control blocks with a single meaningful line,
    suggesting conversion to one-liners (if under line length
    limit), even if nested inside other blocks
    """
    lines   = context["lines"]
    file    = context["file"]
    results = []
    i       = 0
    total   = len(lines)

    while i < total:
        line  = lines[i]
        sline = line.strip()

        i += 1
        if sline.startswith(KEYWORDS) and sline.endswith(":"):
            indent = len(line) - len(line.lstrip())
            block  = []
            j      = i

            while j < total:
                next_line = lines[j]
                if not next_line.strip(): j += 1; continue
                next_indent = len(next_line) \
                            - len(next_line.lstrip())
                if next_indent <= indent: break
                block.append((j, next_line.strip()))
                j += 1

            # Check only meaningful lines
            exclude = list(KEYWORDS) + ["#"]
            body    = [b for _, b in block if b and not
                      (b.startswith(e) for e in exclude)]
            if len(body) == 1:
                compact = f"{line.rstrip()} {body[0]}"
                if len(compact) <= LINE_LEN:
                    results.append({
                        "file": file,
                        "line": i,
                        "col": 1,
                        "code": "Z200",
                        "msg": "control block could be "
                               "compacted to a one-liner"
                    })

    return results



from pathlib import Path

from cnat import api


def translate(path: Path) -> Path | bool:
    found = False

    with open(path) as f:
        for line in f:
            if not line.startswith("#"): continue
            if "@natlang" in line:
                try:
                    lang  = line.split(":::")[1]
                    found = True
                except IndexError: pass

    if not found: return path, False
    return api.cnat(path, lang, get=True), True
    


import importlib.util
import site
import os

from drace.utils import Align, any_eq, all_in, any_in
from drace.types import Context, Dict
from drace.utils import find_proot


PROOT = None


def _module_spec_origin(name: str) -> str | None:
    """
    Try to locate module spec origin. Returns:
      - absolute path to the module file if found
      - the strings 'builtin' for builtins
      - '__future__' for __future__
      - None when spec couldn't be resolved
    """
    if name == "__future__": return name

    parts = name.split(".")
    for i in range(len(parts), 0, -1):
        prefix = ".".join(parts[:i])
        try: spec = importlib.util.find_spec(prefix)
        except Exception: spec = None
        if spec:
            origin   = getattr(spec, "origin", None)
            builtins = ("built-in", "frozen")
            if origin is None or any_eq(builtins, eq=origin):
                if name not in PROOT: return "builtin"

            # Normalize to absolute path if it looks like a
            # filesystem path
            try: origin_abs = os.path.abspath(origin)
            except Exception: origin_abs = origin
            return origin_abs

        # Heuristic for relative imports
        elif name.startswith("."): return PROOT.lower()
    return None


def _is_editable_third_party(path: str) -> bool:
    """Check for .pth files that indicate editable install"""
    if PROOT.lower() in path: return False

    try: pkgs = os.listdir(site.getsitepackages()[0])
    except FileNotFoundError: return False

    editables = [pkg.split(".", 1)[1].replace("pth",
                "dist-info") for pkg in pkgs if all_in(
                "editable", "pth", eq=pkg)]

    for editable in editables:
        for pkg in pkgs:
            if editable == pkg: return True

    return False


def _classify_import(name: str, cwd: str) -> str:
    """
    Return one of: 'FUTURE', 'STANDARDS', 'THIRD_PARTIES',
    'LOCALS'

    Robust against origin values that are special tokens or
    non-absolute paths.
    """
    if name == "__future__": return "FUTURE"

    origin = _module_spec_origin(name)
    # If we couldn't resolve an origin, conservatively call
    # it THIRD_PARTIES
    if not origin: return "THIRD_PARTIES"

    # Handle special tokens
    if origin == "__future__": return "FUTURE"
    if origin == "builtin": return "STANDARDS"

    # At this point origin should be a path-like string
    # (absolute or relative).
    try: origin_abs = os.path.abspath(origin)
    except Exception: origin_abs = origin

    origin_l = origin_abs.lower() if isinstance(origin_abs,
               str) else ""

    # Heuristic: if 'python' appears in the origin path
    # (stdlib) -> STANDARDS
    if "python" in origin_l and not any_in("site-packages",
        "dist-packages", eq=origin_l): return "STANDARDS"

    # site-packages and dist-packages indicate third-party
    if any_in("site-packages", "dist-packages", eq=origin_l)\
            or _is_editable_third_party(origin_l):
        return "THIRD_PARTIES"

    # Fallback: if not stdlib or 3rd-party, then local
    return "LOCALS"


def _import_key_length(line: str) -> int:
    """
    Used to sort imports by descending physical line length.
    """
    return len(line.rstrip("\n"))


def _render_darkian_block(grouped_lines: dict[str,
        list[str]], preserve_order: list[str]) -> str:
    """
    Build the Darkian-standard import block string.

    `grouped_lines` maps group name -> list of import lines
    (strings).
    `preserve_order` ensures FUTURE comes first etc.
    """
    sections = []
    # FUTURE at top if present
    for group in preserve_order:
        lines = grouped_lines.get(group, [])
        if not lines: continue
        if group == "FUTURE":
            sections.append("\n".join(lines))
            sections.append("")  # blank line separator
            continue
        # add section header for non-FUTURE groups
        center = Align(offset=2).center
        if group == "STANDARDS":
            sections.append(f"\n# {center(' STANDARDS ',
                '=')}")
        elif group == "THIRD_PARTIES":
            sections.append(f"# {center(' THIRD PARTIES ',
                '=')}")
        elif group == "LOCALS":
            sections.append(f"# {center(' LOCALS ', '=')}")
        sections.extend(lines)
        sections.append("")  # blank line after section

    # Trim trailing blank lines
    while sections and sections[-1] == "": sections.pop()
    return "\n".join(sections)


def check_z101(context: Context) -> list[Dict]:
    """
    In import blocks, order imports by descending line
    length — grouped by Darkian Standard.
    """
    global PROOT
    lines   = context["lines"]
    file    = context["file"]
    PROOT   = find_proot(file)
    results = []
    cwd     = os.getcwd()

    # Collect contiguous import blocks: sequence of
    # import/from lines (allow inline comments but break on
    # other code)
    import_blocks = []
    cur_block     = []
    cur_start     = None
    for i, raw in enumerate(lines):
        s = raw.strip()
        if s.startswith("import ") or s.startswith("from "):
            if cur_start is None: cur_start = i
            cur_block.append((i, raw))
        else:
            if cur_block:
                import_blocks.append((cur_start, cur_block))
                cur_block = []
                cur_start = None
    if cur_block: import_blocks.append((cur_start, cur_block))

    for i, (start_idx, block) in enumerate(import_blocks):
        # block: list of (index, line)
        # Build classification per import line; also preserve
        # original text
        grouped: dict[str, list[tuple[str, int]]] = {
            "FUTURE": [],
            "STANDARDS": [],
            "THIRD_PARTIES": [],
            "LOCALS": []
        }
        for idx, (li, line) in enumerate(block):
            stripped = line.strip()
            # parse the import name for heuristics:
            # handle 'from X import ...' or 'import X as Y'
            # or 'import X, Y'
            module_name = None
            if stripped.startswith("from "):
                # from X import ...
                parts = stripped.split()
                if len(parts) >= 2: module_name = parts[1]
            elif stripped.startswith("import "):
                # import a, b as c
                rest = stripped[len("import "):].split(","
                       )[0].strip()
                # take first module name, remove 'as ...' if
                # present
                module_name = rest.split()[0]
            if module_name is None: module_name = ""

            grp = _classify_import(module_name, cwd)
            # store original line and its raw length for
            # later sorting by descending length
            grouped.setdefault(grp, []).append((
                line.rstrip("\n"), start_idx + idx + 1))

        # If everything already ordered by descending length
        # within each section and sections in right order,
        # skip
        # Build grouped lines sorted by descending length
        preserve_order = ["FUTURE", "STANDARDS",
                          "THIRD_PARTIES", "LOCALS"]
        grouped_sorted_texts: dict[str, list[str]] = {}
        correct_order = []
        for g in preserve_order:
            items = grouped.get(g, [])
            if not items:
                grouped_sorted_texts[g] = []
                continue
            # sort by descending length of the import text
            sorted_items = sorted(items, key=lambda t:
                           len(t[0]), reverse=True)
            grouped_sorted_texts[g] = [t[0] for t in
                                       sorted_items]
            correct_order.extend(grouped_sorted_texts[g])

        current_order = []
        for _, statement in import_blocks[i][1]:
            current_order.append(statement)

        # If not out of order overall, continue
        if current_order == correct_order: continue

        # Otherwise, produce a suggestion block in Darkian
        # format
        suggestion_text = _render_darkian_block(
                          grouped_sorted_texts,
                          preserve_order)

        # Build Z101 results for every line that is
        # out-of-order (we'll flag the original positions)
        # For clarity give one aggregated result at the start
        # of the block with the suggestion
        results.append({
            "file": file,
            "line": start_idx + 1,
            "col": 1,
            "code": "Z101",
            "msg": "import block not ordered by Darkian "
                   "Standard (descending length per "
                  f"section). Suggestion:{suggestion_text}"
        })

    return results



from drace.linter import engine
from drace.utils import *
from pathlib import Path
import time
import os

SCORE = 0
FILES = 0


def format_cmd(path, diff=False):
    # Placeholder
    print(f"[drace] formatting {path} (diff={diff})")


def lint_cmd(path: str, score: bool, first: bool,
             done: bool = False) -> int:
    results = engine.scrutinize(path)

    mode = "scoring" if "score" in sys.argv else "linting"
    if mode == "linting":
        if not results: return 0
        transmit(f"{mode} {color(pathit(path), WHITE)}\n")
    elif first:
        if not done: path = str(Path(path).resolve().parent)
        path = color(path.split(os.sep)[-1], WHITE)
        transmit(f"{mode} {path}")

    if not results: return 0

    ldeno = len(str(max(r['line'] for r in results)))
    cdeno = len(str(max(r['col'] for r in results)))
    codes = set(r['code'] for r in results)

    act_on = results if mode == "linting" else []
    for r in act_on:
        code  = r['code']
        bold  = code == "E001"
        file  = r['file'].split(os.sep)[-1]
        line  = color(format_order(r['line'], ldeno), YELLOW)
        col   = format_order(r['col'], cdeno)
        ccode = color(code, BAD, bold=bold)
        msg   = r['msg'].strip()

        if code == "Z101":
            msg, rest = msg.split("#", 1)
            check_msg = msg.split(":", 1)
            if len(check_msg) > 1 and check_msg[1] != "":
                msg = check_msg[0] + ":"
                if check_msg[1] != "\n":
                    msg += "\n\n" + check_msg[1].strip()

        first      = f"{file}{SEP}{line}{SEP}{col} {ccode} "
        suggestion = f"{first}{msg}"
        indent     = visual_width(first) if WRAP else 0
        print(wrap_text(suggestion, indent))
        if code == "Z101": print(f"\n#{rest}\n")
        time.sleep(SPEED)

    if score or mode == "scoring":
        score_it([results, done], mode)
    else: print()

    return 1 if "E001" in codes else 0


def score_it(linted: list[list[dict], bool],
             mode: str) -> None:
    global FILES, SCORE
    end        = "\n" if mode == "linting" else ""
    results    = linted[0]
    all_issues = len(results)
    all_lines  = sum(1 for _ in open(results[0]['file']))\
              if results else 1

    score = 100 if all_lines == 0 else 100 \
          * (1 - all_issues / all_lines)

    if linted:
        SCORE += score
        FILES += 1
        if linted[1]: score = SCORE / FILES

    if not linted or linted[1]:
        score = pc_colored(max(0, score))
        transmit(f"code {score} Darkian Standard\n", end=end)



import ast
import re

from drace.types import Context, Dict
from drace.constants import KEYWORDS


_ASSIGN_RE = re.compile(r"(?<![=!<>+\-*/%&|^])=(?!=)")


def _find_assignment_pos(line: str) -> int | None:
    """
    Return column index (0-based) of the assignment '=' for
    real assignments, or None
    """
    m = _ASSIGN_RE.search(line)
    if not m: return None
    return m.start()


def _is_simple_assignment(node: ast.AST) -> bool:
    """
    Return True for top-level Assign or AnnAssign statements
    that are not inside control-flow statements
    """
    if not isinstance(node, (ast.Assign, ast.AnnAssign)):
        return False

    # Walk up the parent chain to ensure it's not inside an
    # `If`, `For`, etc.
    control_structs = (ast.If, ast.For, ast.While, ast.With,
                       ast.Try, ast.Match)
    parent = getattr(node, 'parent', None)
    while parent:
        if isinstance(parent, ast.Module): break
        if isinstance(parent, control_structs): return False
        parent = getattr(parent, 'parent', None)

    return True


def _line_indentation(line: str) -> int:
    """Count leading spaces (or tabs)."""
    return len(line) - len(line.lstrip(' '))


def _is_simple_assignment_line(line: str) -> bool:
    """Naive assignment check, will improve if needed."""
    stripped = line.lstrip()
    # Exclude if line starts with control keyword followed by
    # assignment
    if any(stripped.startswith(kw) for kw in KEYWORDS):
        return False

    return '=' in stripped and not stripped.startswith('#')


def _group_assignments_by_indent(assign_ln:
        list[int], lines: list[str]) -> list[list[int]]:
    groups        = []
    current_group = []
    last_indent   = None

    for i, lineno in enumerate(assign_ln):
        line = lines[lineno - 1]
        if not _is_simple_assignment_line(line): continue

        indent = _line_indentation(line)

        if not current_group:
            current_group.append(lineno)
            last_indent = indent
            continue

        prev_lineno = assign_ln[i - 1]
        if lineno == prev_lineno + 1:
            prev_line = lines[prev_lineno].strip()
            if indent == last_indent and prev_line:
                current_group.append(lineno)
            else:
                groups.append(current_group)
                current_group = [lineno]
                last_indent   = indent
        else:
            groups.append(current_group)
            current_group = [lineno]
            last_indent   = indent

    if current_group: groups.append(current_group)

    return groups


def _collect_assignment_lines_and_blocks(tree
        ) -> list[tuple[int, ast.AST]]:
    """
    Return list of (lineno, node) for assignment-like
    statements found in AST, sorted by lineno.
    """
    out = []

    for node in ast.walk(tree):
        if _is_simple_assignment(node):
            # Some AnnAssign (x: int = 1) may not have lineno
            # if generated; check presence
            if hasattr(node, "lineno"):
                out.append((node.lineno, node))

    out.sort(key=lambda t: t[0])
    return out


def check_z100(context: Context) -> list[Dict]:
    """
    Z100: Enforce vertical alignment of `=` in real
          assignment blocks.
    """
    lines     = context["lines"]
    tree      = context["tree"]
    file      = context["file"]
    results   = []
    assigns   = _collect_assignment_lines_and_blocks(tree)
    assign_ln = [ln for ln, _ in assigns]
    groups    = _group_assignments_by_indent(assign_ln, lines)

    for group in groups:
        # build the lines corresponding to those assignment
        # line numbers
        eq_positions = []
        eq_pos_map   = {}
        for idx, ln in enumerate(group):
            line = lines[ln - 1]
            pos  = _find_assignment_pos(line)
            if pos is None: continue
            eq_positions.append(pos)
            eq_pos_map[ln] = pos
        if not eq_positions: continue
        target = max(eq_positions)
        for ln, pos in eq_pos_map.items():
            if pos != target:
                results.append({
                    "file": file,
                    "line": ln,
                    "col": pos + 1,
                    "code": "Z100",
                    "msg": "assignment not vertically aligned"
                })

    return results



import ast

from drace.types import Context, Dict
from drace import utils


CONTROL_NODE_TYPES = (ast.If, ast.For, ast.While, ast.With,
                      ast.Try, ast.AsyncFor, ast.AsyncWith)


def _is_comment_or_empty(line: str) -> bool:
    s = line.strip()
    return not s or s.startswith("#")


def _is_semicolon_compound(line: str) -> bool:
    """
    Return True if the physical line contains multiple
    independent Python statements
    separated by semicolons. We use
    `utils.tolerant_parse_module` to avoid false positives
    (strings, SQL, etc).
    """
    tree = utils.tolerant_parse_module(line)
    # Treat as compound only if parse produced 2+ top-level
    # statements
    return len(getattr(tree, "body", [])) >= 2


def check_z201(context: Context) -> list[Dict]:
    """
    Z201: Warn against over-compacted blocks.

    Flags:
      - control blocks written on a single physical line whose
        total length > utils.LINE_LEN
      - semicolon-joined single physical lines that parse to
        multiple statements and exceed utils.LINE_LEN

    Conservative: uses AST to avoid false positives.
    """
    lines   = context["lines"]
    tree    = context["tree"]
    file    = context["file"]
    results = []
    src     = "\n".join(lines)
    for node in ast.walk(tree):
        if isinstance(node, CONTROL_NODE_TYPES):
            # Ensure node has location info
            if not hasattr(node, "lineno") \
                or not hasattr(node, "end_lineno"): continue
            # Only care about nodes that occupy a single
            # physical line
            if node.lineno != node.end_lineno: continue
            # Get src for the node; ast.get_source_segment
            # may return None in some cases
            try: node_src = ast.get_source_segment(src, node)\
                         or ""
            except Exception:
                # Fallback: slice from lineno..end_lineno
                start    = node.lineno - 1
                node_src = lines[start].rstrip("\n")
            if not node_src: continue

            col = (getattr(node, "col_offset", 0) + 1) if \
                   hasattr(node, "col_offset") else 1
            if len(node_src) > utils.LINE_LEN:
                results.append({
                    "file": file,
                    "line": node.lineno,
                    "col": col,
                    "code": "Z201",
                    "msg": "overly compact block exceeds "
                           "line length; consider splitting"
                })

    for i, raw in enumerate(lines):
        if _is_comment_or_empty(raw): continue
        # Only consider physically single lines (they already
        # are, since iterating lines)
        if ";" not in raw: continue
        # Quick length check first to avoid repeated parsing
        if len(raw) <= utils.LINE_LEN: continue
        # Try AST parse to verify this line is truly multiple
        # statements
        if _is_semicolon_compound(raw):
            results.append({
                "file": file,
                "line": i + 1,
                "col": 1,
                "code": "Z201",
                "msg": "overly compact block exceeds line "
                       "length; consider splitting"
            })

    return results



import builtins
import ast

from drace.types import Context, Dict
from drace import utils

# =================== Z220 to Z229 Draft ====================

BUILTINS = set(dir(builtins))

# Helper utilities

def _get_assignments(node):
    return [n for n in ast.walk(node) if isinstance(n, (ast.Assign, ast.AnnAssign))]

def _get_function_defs(node):

    return [n for n in ast.walk(node) if isinstance(n, ast.FunctionDef)]

def _hash_node(node: ast.AST) -> str:
    """Return a normalized string representation for repeated logic detection."""

    name_map     = {}
    name_counter = [0]

    def _normalize_name(name):
        if name not in name_map:
            name_map[name] = f"var{name_counter[0]}"
            name_counter[0] += 1
        return name_map[name]

    def _clone(n):
        if isinstance(n, ast.AST):
            new_node = n.__class__()
            for field, value in ast.iter_fields(n):
                if field in ('lineno', 'col_offset', 'end_lineno', 'end_col_offset'):
                    continue
                if isinstance(value, ast.AST) or isinstance(value, list):
                    setattr(new_node, field, _clone(value))
                elif isinstance(value, str):
                    setattr(new_node, field, "_str")
                elif isinstance(value, (int, float, bool)):
                    setattr(new_node, field, "_const")
                elif isinstance(value, ast.AST):
                    setattr(new_node, field, _clone(value))
                else:
                    setattr(new_node, field, value)

            if isinstance(n, ast.Name):
                new_node.id = _normalize_name(n.id)

            return new_node
        elif isinstance(n, list):
            return [_clone(x) for x in n]
        else:
            return n

    cleaned = _clone(node)
    return ast.dump(cleaned, annotate_fields=False, include_attributes=False)


# --------------- Z221: Orthogonal Functions ----------------
def Z221_check(tree, file: str) -> list[dict]:
    results = []

    composite_blocks = (ast.If, ast.For, ast.While, ast.Try, 
                        ast.With, ast.Match)
    for func in _get_function_defs(tree):
        top_blocks = [n for n in func.body if isinstance(
                      n, composite_blocks)]
        steps      = len(top_blocks)
        if steps > utils.MAX_STEPS:
            message = f"function has {steps} top-level " \
                    + "steps review for possible " \
                    + "decomposition if clarity or cohesion"\
                    + " is affected"
            results.append({
                'file': file,
                'line': func.lineno,
                'col': 1,
                'code': 'Z221',
                'msg': message
            })
    return results


# ---------------- Z222: Tight Coupling ----------------
def Z222_check(tree, file: str) -> list[dict]:
    results = []

    for func in _get_function_defs(tree):
        local_vars = {n.id for n in ast.walk(func) if 
                     isinstance(n, ast.Name) and
                     isinstance(getattr(n, 'ctx', None),
                     ast.Store)}
        ignored    = BUILTINS | local_vars | {arg.arg for arg
                     in func.args.args}
        external_objects = set()
        for node in ast.walk(func):
            if isinstance(node, ast.Attribute) and isinstance(node.value, ast.Name):
                name = node.value.id
                if name not in ignored:
                    external_objects.add(name)

        if len(external_objects) > utils.MAX_COUPLING:
            results.append({
                'file': file,
                'line': func.lineno,
                'col': 1,
                'code': 'Z222',
                'msg': f'function uses {len(external_objects)} external objects; may be tightly coupled'
            })

    return results


# -------------- Z223: Implicit State Mutation --------------
def Z223_check(tree, file: str) -> list[dict]:
    results = []

    for func in _get_function_defs(tree):
        for node in ast.walk(func):
            if isinstance(node, ast.Assign) and isinstance(node.targets[0], ast.Attribute):
                attr = node.targets[0]
                if isinstance(attr.value, ast.Name) and attr.value.id not in {arg.arg for arg in func.args.args} | {'self'}:
                    results.append({
                        'file': file,
                        'line': node.lineno,
                        'col': 1,
                        'code': 'Z223',
                        'msg': 'assignment mutates external state implicitly; make explicit'
                    })
    return results


# ---------------- Z224: Parameter Explosion ----------------
def Z224_check(tree, file: str) -> list[dict]:
    results = []

    for func in _get_function_defs(tree):
        if len(func.args.args) > 6:
            results.append({
                'file': file,
                'line': func.lineno,
                'col': 1,
                'code': 'Z224',
                'msg': f'function has {len(func.args.args)} parameters; consider grouping or refactoring'
            })
    return results


# ------------ Z225: Overloaded Data Structures -------------
def Z225_check(tree, file: str) -> list[dict]:
    results = []

    for node in ast.walk(tree):
        # --- Skip dicts used in .items() pattern ---
        if isinstance(node, ast.Dict) and len(node.keys) > 5:
            parent = getattr(node, 'parent', None)
            if isinstance(parent, ast.Call) and isinstance(
                    parent.func, ast.Attribute):
                if parent.func.attr == 'items': continue

            results.append({
                'file': file,
                'line': node.lineno,
                'col': 1,
                'code': 'Z225',
                'msg': 'dict has too many keys; consider splitting responsibilities'
            })

        elif isinstance(node, ast.ClassDef):
            public_attrs = [
                n.targets[0].id for n in node.body
                if isinstance(n, ast.Assign)
                and isinstance(n.targets[0], ast.Name)
            ]
            if len(public_attrs) > 5:
                results.append({
                    'file': file,
                    'line': node.lineno,
                    'col': 1,
                    'code': 'Z225',
                    'msg': 'class has many public attributes; consider SRP'
                })

    return results


# ---------------- Z226: Temporal Coupling ----------------
def Z226_check(lines: list[str], file: str) -> list[dict]:
    # Simplified heuristic: look for consecutive assigns to same variable across lines
    results = []
    last_assigned = {}
    for i, line in enumerate(lines):
        tree = utils.tolerant_parse_module(line)
        if not tree.body: continue
        node = tree.body[0]
        if isinstance(node, ast.Assign) and isinstance(node.targets[0], ast.Name):
            var = node.targets[0].id
            if var in last_assigned and i - last_assigned[var] <= 2:
                results.append({
                    'file': file,
                    'line': i + 1,
                    'col': 1,
                    'code': 'Z226',
                    'msg': f'variable {var} assigned in temporal sequence; possible temporal coupling'
                })
            last_assigned[var] = i
    return results


# ---------------- Z228: Abstraction Leak ----------------
def Z228_check(tree, file: str) -> list[dict]:
    results = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Return):
            if isinstance(node.value, (ast.List, ast.Dict, ast.Set)):
                results.append({
                    'file': file,
                    'line': node.lineno,
                    'col': 1,
                    'code': 'Z228',
                    'msg': 'returning internal data structure; consider interface or copy'
                })
    return results


# --------------- Z229: Unstable API Patterns ---------------
def Z229_check(tree, file: str) -> list[dict]:
    results = []

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            returns = [n for n in ast.walk(node) if isinstance(n, ast.Return)]
            types_seen = set()
            for r in returns:
                if hasattr(r.value, '__class__'):
                    types_seen.add(type(r.value).__name__)
            if len(types_seen) > 3:
                results.append({
                    'file': file,
                    'line': node.lineno,
                    'col': 1,
                    'code': 'Z229',
                    'msg': 'function has multiple return types; consider consistent API'
                })
    return results

# ------------------------ Run Checks -----------------------
def check_Z22_(context: Context) -> list[Dict]:
    rules = [Z221_check, Z222_check, Z223_check, Z224_check,
             Z225_check, Z228_check, Z229_check]

    lines   = context["lines"]
    tree    = context["tree"]
    file    = context["file"]
    results = []

    for rule in rules: results.extend(rule(tree, file))
    results.extend(Z226_check(lines, file))

    return results



"""API for running pyflakes."""
# ======================== STANDARDS ========================
from io import StringIO

# ========================== LOCALS =========================
from drace.utils import tolerant_parse_module
from .reporter import Reporter
from .checker import Checker


__all__ = ['check']


def check(filename: str, buffer: StringIO):
    """
    Check the Python source given by C{source} for flakes.
    """
    with open(filename) as f: source = f.read()
    reporter = Reporter(buffer, buffer)

    # First, compile into an AST and handle syntax errors.
    tree = tolerant_parse_module(source)

    # Okay, it's syntactically valid.  Now check it.
    w = Checker(tree, filename=filename)
    w.messages.sort(key=lambda m: m.line)
    for warning in w.messages: reporter.flake(warning)



# ======================= STANDARDS =========================
from pathlib import Path
import io

# ========================= LOCALS ==========================
from drace.constants import IGNORED_RULES, ONLY
from drace.darkian import get_rules
from .pycodestyle import Checker
from .pyflakes import flake_api
from drace import utils


IGNORE = ("E113", "E121", "E124", "E126", "E127", "E128", 
          "E131", "E221", "E222", "E701", "E702", "E704") \
       + tuple(IGNORED_RULES)


def run_style_checks(file: str | Path) -> list[dict]:
    """Run Darkian-patched pycodestyle checks on a file."""
    file = str(file)

    # Build checker and capture final
    checker = Checker(file)
    checker.check_all()

    results = []
    for line, col, code, msg in checker.report.errors:
        if ONLY and code not in ONLY: continue
        if code in IGNORE: continue
        results.append({
            "file": file,
            "line": line,
             "col": col,
            "code": code,
             "msg": " ".join(msg.strip().split()[1:])
        })

    return results


def run_flake_checks(file: str | Path) -> list[dict]:
    """Run pyflakes checks on a file."""
    def format_flake(msg: str) -> tuple[str]:
        code = "Z999"
        if "imported but unused" in msg: code = "W611"
        if "star imports" in msg:
            code  = "F405"
            parts = msg.split(":")
            stars = parts[1].split(",")
            imp   = "imports:" if len(stars) > 1 else "import:"
            s_str = ""
            for i, star in enumerate(stars):
                sep = "," if s_str else ""
                if i and i == len(stars) - 1: sep = " or"
                s_str += f"{sep} {star.strip()}"
            msg = parts[0].replace("imports", imp) + s_str
        if "syntax error" in msg.lower(): code = "E001"
        if "*' used; unable t" in msg: code = "F403"
        if "undefined name" in msg: code = "E602"
        if "but never used" in msg: code = "W612"
        if "f-string is" in msg: code = "F541"

        return code, msg

    buffer = io.StringIO()
    flake_api.check(str(file), buffer)

    results = []
    for warning in buffer.getvalue().splitlines():
        parts = warning.split(":", 3)[1:]
        if len(parts) == 3:
            line, col, _  = parts
            code, message = format_flake(_)
            if ONLY and code not in ONLY: continue
            if code in IGNORE: continue
            if "unexpected indent" in message: continue
            if "unterminated stri" in message: continue
            results.append({
                "file": file,
                "line": int(line.strip()),
                 "col": int(col.strip()),
                "code": code,
                 "msg": message
            })

    return results


def run_darkian_checks(file: str | Path) -> list[dict]:
    """Run Darkian checks on a file."""
    lines   = Path(file).read_text(encoding="utf-8")\
              .splitlines()
    file    = str(file)
    results = []

    tree, synerrs = utils.tolerant_parse_module(lines, True)
    utils.annotate_parents(tree)
    tree.parent = tree

    context = {
        "lines": lines,
         "tree": tree,
         "file": file,
    }

    for rule in get_rules(IGNORE, ONLY):
        results.extend(rule(context))

    for synerr in synerrs:
        if "triple-" in synerr[1]: continue
        if "invalid" in synerr[1]: continue
        if "string " in synerr[1]: continue

        results.append({
            "file": file,
            "line": synerr[0],
             "col": 1,
            "code": "E001",
             "msg": f"Syntax Error: {synerr[1]}"
        })

    return results


def scrutinize(file: str | Path) -> list[dict]:
    """Lint a file."""
    final  = []
    final += run_style_checks(file)
    final += run_flake_checks(file)
    final += run_darkian_checks(file)
    return sorted(final, key=lambda x: (x["line"], x["col"]))



"""
Tests for detecting redefinition of builtins.
"""
from pyflakes import messages as m
from pyflakes.test.harness import TestCase


class TestBuiltins(TestCase):

    def test_builtin_unbound_local(self):
        self.flakes('''
        def foo():
            a = range(1, 10)
            range = a
            return range

        foo()

        print(range)
        ''', m.UndefinedLocal)

    def test_global_shadowing_builtin(self):
        self.flakes('''
        def f():
            global range
            range = None
            print(range)

        f()
        ''')



# ======================= STANDARDS =========================
from typing import Callable
from pathlib import Path
import importlib
import os

# ========================= LOCALS ==========================
from drace.types import Context, Dict


def get_rules(ignore: tuple[str], only: list[str]
             ) -> list[Callable[[Context], list[Dict]]]:
    """
    Dynamically discover and load all callable linting rule
    functions from the 'drace/darkian' subdirectories
    (organized by series).

    Each rule is expected to:
      - Reside in a .py file named descriptively (excluding
        underscores or dot-prefixed).
      - Contain a function starting with `check_`.
      - Return a list of Dicts when called with a Context.

    Args:
        ignore: A tuple of rule names (uppercase, no 'check_'
                prefix) to exclude.
        only: A list of specific rule names to load. If
              non-empty, all others are skipped.

    Returns:
        A list of functions that take a Context and return a
        list of Dicts.
    """
    rules   = []
    pvt     = ["_", "."]
    pkg     = __package__
    darkian = os.listdir(Path(__file__).resolve().parent)
    series  = [s for s in darkian if not any(s.startswith(c)
              for c in pvt)]

    for s in sorted(series):
        path = Path(__file__).resolve().parent / Path(s)
        for rule in os.listdir(path):
            # Skip private or non-.py files
            if any(rule.startswith(c) for c in pvt): continue
            name = rule[:-3]  # Strip '.py' extension

            # Filter out unwanted rules based on name
            if only and name.upper() not in only: continue
            if name.upper() in ignore: continue

            # Dynamically import the module and look for
            # functions starting with 'check_' in the module
            mod = importlib.import_module(f"{pkg}.{s}.{name}")
            for attr in dir(mod):
                if attr.startswith("check_"):
                    fn = getattr(mod, attr)
                    if callable(fn): rules.append(fn)

    return rules
